
/* $Id: pvmrun.c,v 1.14 2005/06/22 17:42:39 amit Exp $ */

static char rcsid[] = "$Id: pvmrun.c,v 1.14 2005/06/22 17:42:39 amit Exp $";

/*
 
 A wrapper program that creates a PVM virtual machine based on a list of
 machines supplied by PBS. It allows spawning either a conventional
 master/slave style pvm program or a spmd style pvm program. It is also
 convenient for running spmd style PVM programs without using a console,
 if so desired. Easily modified to use with other workload management systems.

 Copyright: Amit Jain, 2003.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



*/

/*
	$Log: pvmrun.c,v $
	Revision 1.14  2005/06/22 17:42:39  amit
	Add license to the program.

	Revision 1.13  2004/11/30 03:55:40  amit
	Fix working directory for non PBS version. Make pvmrun to use
	current working directory without the use of PVM_WD directory.

	Revision 1.12  2004/11/30 03:34:11  amit
	Add handling of working directory for spmd programs. Earlier
	PVM_WD only worked for master slave programs.

	Revision 1.11  2004/10/10 03:56:50  amit
	Further changes to make single machine mode work better.

	Revision 1.10  2004/09/27 03:04:20  amit
	Revert np=1 to default behavior of running on master node.

	Revision 1.9  2004/09/24 13:45:29  amit
	Make sure that spmd jobs do not spawn on the master node

	Revision 1.8  2004/09/02 00:00:28  amit
	Change spawning so that master runs on the master node.

	Revision 1.7  2004/06/22 21:21:33  amit
	Further changes to deal with PBS properly.

	Revision 1.6  2004/06/11 05:33:55  amit
	Improve help message.

	Revision 1.5  2004/06/11 05:31:03  amit
	Change help message for PBS only case.

	Revision 1.4  2004/06/11 05:23:16  amit
	Add support for PBS only execution.

	Revision 1.3  2003/11/17 23:48:17  amit
	Add help messages.

	Revision 1.2  2003/11/17 21:53:12  amit
	Added signal handling to terminate PVM if user interrupts.
	Added check for hostfile to be found in current, home directories.
	Set default hostfile to /usr/local/etc/machines (same as for MPI)

	Revision 1.1  2003/11/17 08:52:28  amit
	Initial revision

*/

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>  // must include stdio  for pvm3.h to compile correctly
#include <pvm3.h>

const int TASK_DONE = 1;
const int TRUE=1;
const int FALSE=0;
const int MAXSIZE=1024;
const int MAX_HOSTNAME_LENGTH=256;
const int DEBUG=2;

char *program_name;
char *nodefile; 
char *DEFAULT_NODEFILE = "/usr/local/etc/machines";

#define BATCH_SYSTEM "PBS"
#define BATCH_NODEFILE "/var/spool/pbs"
#define BATCH_NODEFILE_LENGTH 14
#define BATCH_SYSTEM_ONLY TRUE

void configure_pvm(int, char **);
void stop_pvm(char *);
char *check_node_file(void);
void start_pvm_system(char **);
char **read_node_file(int, char *);
void spawn_pvm_tasks(char *, char **, int , int *);
int wait_for_tasks_to_exit(int , int *);
void print_usage();
void print_help();
void sig_handler(int signo);
void setup_signal_handling();


void setup_signal_handling()
{
	/* catch interrupt and terminate signals to stop pvm cleanly */
	signal(SIGINT, sig_handler);
	signal(SIGTERM, sig_handler);
}


void sig_handler(int signo)
{
	switch (signo) {
		case SIGINT:
		case SIGTERM:
			stop_pvm("Caught interrupt. Halting PVM.");
			exit(1);
			break;
		default:
			break;
	}
}

void stop_pvm(char *message)
{
	fprintf(stderr, "%s: %s.\n", program_name, message);
	signal(SIGTERM, SIG_IGN);
	pvm_halt();
	signal(SIGTERM, SIG_DFL);
}


char *check_node_file()
{
	char *nodefile;
	char *pbsfile;
	struct stat buf;
	struct passwd *entry;
	int n;

	nodefile = (char *) malloc(sizeof(char)*MAXSIZE);

    pbsfile = getenv("PBS_NODEFILE");
	if (pbsfile)
	{
		/* check for validity of the batch system nodefile */
		if (strncmp(pbsfile,BATCH_NODEFILE,BATCH_NODEFILE_LENGTH) == 0)
		{
			n = strlen(pbsfile);
			strncpy(nodefile, pbsfile, n+1);
			return nodefile;
		} else {
			fprintf(stderr,"\n%s: Invalid PBS_NODEFILE!\n\n", program_name);
			exit(1);
		}
	} else {
		if (BATCH_SYSTEM_ONLY) {
			fprintf(stderr,"\n%s: Must use PBS to get nodes before running pvmrun!\n\n", 
						program_name);
			exit(1);
			
		}
		/* check if .pvm_hosts exists in the current directory */
		strcpy(nodefile, "./.pvm_hosts");
		if (stat(nodefile, &buf) == 0) {
			return nodefile;
		} else { 
			perror("");
			/* try .pvm_hosts in the user's home directory next */
			entry = getpwuid(getuid());
			n = strlen(entry->pw_dir);
			strcpy(nodefile, entry->pw_dir);
			strcat(nodefile, "/.pvm_hosts");
			if (stat(nodefile, &buf) == 0) {
				return nodefile;
			} else {
				/* use systemwide default nodefile /usr/local/etc/machines */
				return DEFAULT_NODEFILE;
			}
		}
	}
	return NULL; /* should never reach here */
}

static int get_num_lines(nodefile)
   char *nodefile;
{
        FILE *fin;
        int num;
        int status;
        char next[MAX_HOSTNAME_LENGTH+1];

        fin = fopen(nodefile, "r");
        if (fin  == NULL)
        {
                fprintf(stderr,"%s: Fatal error: Could not read node file.\n", program_name);
                fprintf(stderr, " %s: Node file path: %s\n", program_name, nodefile);
				exit(1);
        }
        num=0;
        while (!feof(fin)) {
                status = fscanf(fin,"%s", next);
                num++;
        }
        fclose(fin);
        return num-1;
}


char **read_node_file(int nhosts, char *nodefile)
{
	FILE *fin;
	char **hosts;
	int i;
	int status;

	
	hosts = (char **) malloc(sizeof(char *)*nhosts);

	fin = fopen(nodefile, "r");
	if (fin  == NULL)
	{
		fprintf(stderr, "%s: Fatal error: Could not read node file\n", program_name);
		fprintf(stderr, "%s: Node file path: %s\n", program_name, nodefile);
		exit(1);
	}

	if (DEBUG >= 2) {
		fprintf(stderr,"Selected nodes from file: %s\n", nodefile);
		fprintf(stderr,"*************************************************\n");
	}

	for (i=0; i<nhosts; i++) {
		hosts[i] = (char *) malloc(sizeof(char)*MAX_HOSTNAME_LENGTH);
		status = fscanf(fin,"%s", hosts[i]);
		if (DEBUG >= 2) {
			fprintf(stderr,"%s\n", hosts[i]);
		}
		
		if (status == EOF) {
			fprintf(stderr,"%s: Fatal error: Not enough nodes were assigned.\n",
						program_name);
			fprintf(stderr, "%s: Perhaps an incorrect argument to the %s.\n", 
						program_name, BATCH_SYSTEM);
			exit(1);
		}
	}
	if (DEBUG >= 2) {
		fprintf(stderr,"*************************************************\n");
	}
	fclose(fin);
	return hosts;
}

void start_pvm_system(char **argv)
{
	int status;
	sigset_t newmask, oldmask;

	setup_signal_handling(); /* to stop PVM in case of a interrupt */

	/* block interrupts while trying to start pvm */

    sigemptyset(&newmask);
    sigaddset(&newmask, SIGINT);
    sigaddset(&newmask, SIGTERM);
    sigaddset(&newmask, SIGTSTP);
    if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
        fprintf(stderr,"%s: Error in blocking signals\n", program_name);


	fprintf(stderr, "\n");
	fprintf(stderr, "%s: Probing for pre-existing PVM.\n", program_name);
	fprintf(stderr, "***********************************************************\n");
	status = pvm_mytid();
	fprintf(stderr, "***********************************************************\n");

	if (status >= 0)
	{
		stop_pvm("Halting pre-existing PVM");
		fprintf(stderr, "%s: Restart.\n", program_name);
		/* Need to make a clean restart or the rest of the program 
		   stays connected to the non-existent daemon */
		execvp(argv[0], argv);
		exit(1);
	} else {
		fprintf(stderr, "%s: PVM not running. Ignore error messages above from pvm_mytid().\n", program_name);
		fprintf(stderr, "\n");
	}

	status  = pvm_start_pvmd(0, 0, 0);
    if (status  == PvmDupHost) {
		fprintf(stderr,"%s:  Warning: PVM daemon is already running.\n", program_name);
    } else if (status  == PvmSysErr) {
		fprintf(stderr,"%s: Error: Failed to start PVM daemon! Bailing out!!\n", program_name);
        exit(1);
    } else {
		fprintf(stderr,"%s: Succeeded in starting new PVM daemon!\n", program_name);
		fprintf(stderr, "\n");
    }

    /* give the PVM daemon enough time to start up */
    sleep(2);

	/* ok, unblock signals now that PVM is up */
    if (sigprocmask(SIG_UNBLOCK, &newmask, &oldmask) < 0)
	    fprintf(stderr,"%s: Error in unblocking signals\n", program_name);

}


void configure_pvm(int nhosts, char **hosts)
{
	int *info;
	int num_hosts;
	int narch;
	struct pvmhostinfo *hostp;
	int i;
	int status;

    info = (int *) malloc (sizeof(int)*nhosts);
    status = pvm_addhosts(hosts, nhosts, info);
    if (status < 0)
		fprintf(stderr, "%s: Error in trying to add hosts.\n", program_name); 
    else if (status != nhosts) {
		fprintf(stderr, "%s: Succeeded in adding %d hosts.\n", program_name,  status+1);
    }

    pvm_config( &num_hosts, &narch, &hostp );
	fprintf(stderr, "%s: Parallel Virtual Machine configuration\n", 
				program_name);
	fprintf(stderr, "-----------------------------------------------\n"); 
    for (i=0; i<num_hosts; i++) {
		fprintf(stderr, "%s\n", hostp[i].hi_name);
    }
	fprintf(stderr, "-----------------------------------------------\n"); 
	fprintf(stderr, "\n");

}

void spawn_pvm_tasks(char *prog, char **args, int nproc, int *tids)
{
	int numt;
	int i;
	int status;
	int size;
	char *where;
	char *buf1, *buf2;

	/* add working directory option */
  	size = pathconf("/home", _PC_PATH_MAX);
	buf1 = (char *)malloc(sizeof(char)*(size+2));
	buf2 = (char *)malloc(sizeof(char)*(size+2));
	where = (char *)malloc(sizeof(char)*(size+2));
	strcpy(buf1,".:");
	strcpy(buf2,":");
	getcwd(where, size);
	fprintf(stderr, "pvmrun: using working directory %s\n", where);
	strcat(buf1, where);
	strcat(buf2, where);

    pvm_setopt(PvmShowTids, TRUE);
    status = pvm_catchout(stdout);
	if (nproc == 1) {
		numt = pvm_spawn(prog, args, PvmTaskHost, buf1, nproc, tids); 
	} else {
    	if (BATCH_SYSTEM_ONLY) {
			numt = pvm_spawn(prog, args, PvmTaskHost|PvmHostCompl, buf1, nproc, tids); 
		} else {
			numt = pvm_spawn(prog, args, PvmTaskDefault, buf2, nproc, tids); 
		}
	}
	free(buf1);
	free(buf2);
	free(where);

  	if (numt < nproc)
    {
		pvm_perror("Error in spawning nodes");
		fprintf(stderr, "%s: Trouble spawning slaves. Error codes are: \n", 
						program_name);
    	for (i=numt; i<nproc; i++)
		{
			fprintf(stderr,  "TID  %d %d \n", i, tids[i]);
		}
    	for (i=0; i<numt; i++)
      		pvm_kill(tids[i]);
    	pvm_exit();
		stop_pvm("Halting PVM.");
    	exit(1);
    }
}

int wait_for_tasks_to_exit(int nproc, int *tids)
{
	int *done;
	int count;
	int i,j;
	int taskid;

	done = (int *) calloc(nproc, sizeof(int));
    for (i=0; i<nproc; i++)
        done[i] = tids[i];

    pvm_notify(PvmTaskExit, TASK_DONE, nproc, tids);

    /* now wait for notifications */
    for (i=0; i<nproc; i++) {
        pvm_recv(-1, TASK_DONE);
        pvm_upkint(&taskid, 1, 1);
		if (DEBUG >= 2) {
        	fprintf(stderr, "Received PvmTaskExit notification from task id %x\n", taskid);
		}
        for (j=0; j<nproc; j++)
            if (done[j] == taskid) done[j] = -1;
    }

    count = 0;
	for (i=0; i<nproc; i++) {
		if (done[i] == -1) {
			count++;
		}
	}
	return count;
}

void print_usage()
{
	fprintf(stderr,"%s: %s\n", program_name, rcsid);
	fprintf(stderr,"Usage: %s -np <#tasks> <executable> {<args>,...}\n",
	program_name );
}

void print_help()
{
	fprintf(stderr, "************************************************************************\n");
	fprintf(stderr, 
	"  For a pvm master/slave program: %s -np 1 <executable> {<args>,...}\n", program_name);
	fprintf(stderr, 
	"  For a pvm n-way spmd program: %s -np n <executable> {<args>,...}\n", program_name);
	fprintf(stderr, "\n");
	fprintf(stderr, "  The list of machines to use for PVM is determined as follows:\n");
	fprintf(stderr, "    if PBS_NODEFILE is set, then use the file specified by PBS\n");
    if (BATCH_SYSTEM_ONLY) {
		fprintf(stderr, "    else: exit and ask user to run after getting nodes using pbsget.\n");
	} else {
		fprintf(stderr, "    else: first check for  .pvm_hosts in the current directory\n");
		fprintf(stderr, "          then check for .pvm_hosts in the user's home directory\n");
		fprintf(stderr, "          then at last use /usr/local/etc/machines.\n");
		fprintf(stderr, "************************************************************************\n");
	}

}

int main(int argc, char **argv)
{
	int nproc;
	int nhosts;
	char *prog;
	char **prog_args;
	char **hosts;
	int *tids;
	int result;
	char *nodefile;

	program_name = argv[0];
  	if (argc < 4) {
		print_usage();
		print_help();
		exit(1);
	} 

	if (strncmp(argv[1],"-h",2) == 0) {
		print_usage();
		print_help();
		exit(0);
	}

	nproc = atoi(argv[2]);
	prog = argv[3];
	prog_args = argv+4;
	
	nodefile = check_node_file();
	if (DEBUG >= 1) {
		fprintf(stderr, "%s: Using host file: %s \n", program_name, nodefile); 
	}
	start_pvm_system(argv);
	/* use PBS nodefile ro configure PVM */
	nhosts = get_num_lines(nodefile);
	hosts = read_node_file(nhosts, nodefile);
	configure_pvm(nhosts, hosts);

	tids = (int *) calloc(nproc, sizeof(int));
	spawn_pvm_tasks(prog, prog_args, nproc, tids);
	result = wait_for_tasks_to_exit(nproc, tids);
	fprintf(stderr, "\n");
			
	if (result < nproc) {
		fprintf(stderr, "%s: %d tasks spawned but only %d finished!\n", 
					argv[0], nproc,	result);
  		pvm_exit();
		stop_pvm("Halting PVM");
		exit(1);
	} else {
		if (nproc == 1) 
			fprintf(stderr, "%s: master task finished!\n", program_name); 
		else 
			fprintf(stderr, "%s: %d tasks finished!\n", program_name, result); 
	}

  	/* Program finished. Exit PVM */
  	pvm_exit();
	stop_pvm("Halting PVM");
	fprintf(stderr, "%s: End of PVM application.\n", program_name);
  	exit(0);
}

/* vim: set ts=4: */
